/*
 * Licensed to meast-summarize under one or more contributor
 * license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright
 * ownership. IlluSnip licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package ws.meast.summarization.mpsimp;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import name.dxliu.associations.AssociationNode;
import name.dxliu.associations.AssociationTree;
import ws.meast.summarization.mps.ClassDistance;
import ws.meast.summarization.mps.PatternDistance;

 /**
  * 
  * @author Yu Gu
  * This class uses an AStar manner to calculate distance between to association/pattern tree.
  */
public class AStarPatternDistanceImp implements PatternDistance {  
	
	private ClassDistance classDistance;
	private boolean DEBUG_MODE = false;
	public AStarPatternDistanceImp(){

	}
	public void DEBUG(){
		this.DEBUG_MODE = true;
	}
	
	public AStarPatternDistanceImp(ClassDistance cdis){
		this.classDistance = cdis;
	}
	
	public class Mapping{
    	public AssociationNode fromG1;
    	public AssociationNode fromG2;
    	
    	public Mapping(AssociationNode fromG1, AssociationNode fromG2){
    		this.fromG1 = fromG1;
    		this.fromG2 = fromG2;
    	}
    }
	public class Node  
    {         
        public Node(ArrayList<Mapping> editlist)  
        {  
            this.editlist = editlist;  
        }  
        public ArrayList<Mapping> editlist = new ArrayList<Mapping>();  
        public ArrayList<AssociationNode> nodesUnmapped = new ArrayList<AssociationNode>();
        public Node parent;  
          
        public double GE = 0;
        public double GN = 0;
        public double H = 0;  
        public double F = 0;  
          
        public boolean ifcompleted(int G1Size)  
        {  
            return (this.editlist.size() == G1Size);
        }  
    }  
      
    /** 
     * Open Set 
     */  
    public ArrayList<Node> OpenList = new ArrayList<Node>();  
    public ArrayList<Node> CloseList = new ArrayList<Node>();  
    
    public double getDistance(List<Integer> queryEntities, AssociationTree tree1, AssociationTree tree2){
    	double distance = 0;
		Iterator<AssociationNode> itan1 = tree1.getDFSIterator();
		Iterator<AssociationNode> itan2 = tree2.getDFSIterator();
		ArrayList<AssociationNode> Searchings1 = new ArrayList<AssociationNode>();
		ArrayList<AssociationNode> Searchings2 = new ArrayList<AssociationNode>();
		ArrayList<AssociationNode> G1Nodes = new ArrayList<AssociationNode>();
		ArrayList<AssociationNode> G2Nodes = new ArrayList<AssociationNode>();
		for(;itan1.hasNext();){
			AssociationNode node = itan1.next();
			if(queryEntities.contains(node.getId())){
				Searchings1.add(node);
			}
			else
				G1Nodes.add(node);
		}
		for(;itan2.hasNext();){
			AssociationNode node = itan2.next();
			if(queryEntities.contains(node.getId())){
				Searchings2.add(node);
			}
			else
				G2Nodes.add(node);
		}
		
		distance = calculate(Searchings1, Searchings2, G1Nodes, G2Nodes);
    	return distance;
    }

    public double getDistance(List<Integer> queryEntities,List<Integer> codesIn1,List<Integer> codesIn2){

    	double distance = 0;
    	AssociationTree tree1 =AssociationTree.newInstance(AssociationNode.constructViaCanonicalCode(codesIn1));
		Iterator<AssociationNode> itan1 = tree1.getDFSIterator();
	  	AssociationTree tree2 =AssociationTree.newInstance(AssociationNode.constructViaCanonicalCode(codesIn2));

		Iterator<AssociationNode> itan2 = tree2.getDFSIterator();
		ArrayList<AssociationNode> Searchings1 = new ArrayList<AssociationNode>();
		ArrayList<AssociationNode> Searchings2 = new ArrayList<AssociationNode>();
		ArrayList<AssociationNode> G1Nodes = new ArrayList<AssociationNode>();
		ArrayList<AssociationNode> G2Nodes = new ArrayList<AssociationNode>();
		for(;itan1.hasNext();){
			AssociationNode node = itan1.next();
			if(queryEntities.contains(node.getId())){
				Searchings1.add(node);
			}
			else
				G1Nodes.add(node);
		}
		for(;itan2.hasNext();){
			AssociationNode node = itan2.next();
			if(queryEntities.contains(node.getId())){
				Searchings2.add(node);
			}
			else
				G2Nodes.add(node);
		}
		int max = (G1Nodes.size()>=G2Nodes.size())?2*(G1Nodes.size() + Searchings1.size() - 1) + G2Nodes.size():2*(G2Nodes.size() + Searchings2.size() - 1) + G1Nodes.size();
		distance = calculate(Searchings1, Searchings2, G1Nodes, G2Nodes);
		double normDistance = Math.sqrt(distance/max);
    	return normDistance;
    }
    /** 
     * @param Searchings1  list of query_entity in G1. 
     * @param Searchings2 list of query_entity in G2.
     * @param G1Nodes  list of non_query_entity in G1.
     * @param G2Nodes  list of non_query_entity in G2.
     * @return routine.
     */
    public double calculate(ArrayList<AssociationNode> Searchings1,ArrayList<AssociationNode> Searchings2, ArrayList<AssociationNode> G1Nodes, ArrayList<AssociationNode> G2Nodes)  
    {                 
    	//classDis = new DBPediaCooccurrenceClassDistance();
    	OpenList = new ArrayList<Node>();  
    	CloseList = new ArrayList<Node>();  
    	Node start = Initialize(Searchings1,Searchings2);  
    	start.nodesUnmapped.addAll(G2Nodes);
    	int searchsize = Searchings1.size();
        int G1Size = searchsize + G1Nodes.size();          
        Node node = start;  
        Node optnow;
        if(start.ifcompleted(G1Size))
        	optnow = node;       
        else{
        	CloseList.add(node);  
            OpenList.add(Extend(searchsize,Searchings2, G1Nodes, G2Nodes,node,null));
            for(int i = 0; i < G2Nodes.size(); i ++){
            	OpenList.add(Extend(searchsize,Searchings2, G1Nodes, G2Nodes, node, G2Nodes.get(i)));
     
            }
     
            for(optnow = MinF(OpenList); ; optnow = MinF(OpenList)){
            	if(optnow.ifcompleted(G1Size))
            		break;
            	else{
            		OpenList.remove(optnow);
            		OpenList.add(Extend(searchsize,Searchings2, G1Nodes, G2Nodes, optnow, null));
            		for(int i = 0; i < optnow.nodesUnmapped.size(); i ++){
            			OpenList.add(Extend(searchsize,Searchings2, G1Nodes, G2Nodes,optnow,optnow.nodesUnmapped.get(i)));
            		}
            	}
            }
        }
        if(DEBUG_MODE){
        	System.out.println("GE: " + optnow.GE);
        	System.out.println("GN: " + optnow.GN);
        }

        return optnow.F;
    }  
    public void PrintMapping(Node node){
    	ArrayList<Mapping> el = node.editlist;
    	for(int i = 0; i < el.size(); i ++)
    		System.out.println(el.get(i).fromG1.getId() + "," + el.get(i).fromG2.getId());
    }
    private Node Extend(int searchsize, ArrayList<AssociationNode> Searchings2, ArrayList<AssociationNode> G1Nodes, ArrayList<AssociationNode> G2Nodes, Node parent, AssociationNode mapnode){
    	int G1Size = searchsize + G1Nodes.size();
    	ArrayList<Mapping> el = new ArrayList<Mapping>();
    	el.addAll(parent.editlist);
    	int index = parent.editlist.size() - searchsize;
    	Mapping map = new Mapping(G1Nodes.get(index), mapnode);
    	el.add(map);
    	Node node = new Node(el);
    	node.nodesUnmapped.addAll(parent.nodesUnmapped);
    	if(mapnode != null){
    		node.GN = parent.GN + NodeEditCost(map.fromG1.getId(),map.fromG2.getId());
    		node.nodesUnmapped.remove(mapnode);
    	}
    	else
    		node.GN = parent.GN + NodeEditCost(map.fromG1.getId(), 0);
    	node.GE = CalculateEdge(parent, map);
    	node.H = CalculateH(searchsize,G1Nodes,node);
    	node.F = node.GE + node.GN + node.H;
    	node.parent = parent;
    	if(node.ifcompleted(G1Size)){
    		node.F += Withdraw(G2Nodes, Searchings2, node);
    	}
    	return node;
    }
    private double Withdraw(ArrayList<AssociationNode> G2Nodes, ArrayList<AssociationNode> Searchings2, Node node){
    	int edgesleft = 0;
    	for(int i = 0; i < G2Nodes.size(); i ++){
        	boolean flag = false;
        	for(int j = 0; j < node.nodesUnmapped.size(); j ++){
        		if(G2Nodes.get(i).getRelation() != 0 && G2Nodes.get(i).getFather() == node.nodesUnmapped.get(j)){
        			flag = true;
        			break;
        		}
        	}
        	if(flag)
        		edgesleft += 1;
        }
        for(int i = 0; i < Searchings2.size(); i ++){
        	boolean flag = false;
        	for(int j = 0; j < node.nodesUnmapped.size(); j ++){
        		if(Searchings2.get(i).getRelation() != 0 && Searchings2.get(i).getFather() == node.nodesUnmapped.get(j)){
        			flag = true;
        			break;
        		}
        	}
        	if(flag)
        		edgesleft += 1;
        }
        for(int i = 0; i < node.nodesUnmapped.size(); i ++){
        	if(node.nodesUnmapped.get(i).getRelation() != 0)
        		edgesleft += 1;
        }
        int nodesleft = node.nodesUnmapped.size();
        
        return (double)nodesleft + (double)edgesleft;
    }
    private double CalculateH(int searchsize, ArrayList<AssociationNode> G1Nodes, Node node){
    	int G1Size = searchsize + G1Nodes.size();

    	if(node.ifcompleted(G1Size))
    		return 0;
    	int index = node.editlist.size() - searchsize;
    	ArrayList<AssociationNode> nu = new ArrayList<AssociationNode>();
    	ArrayList<AssociationNode> nodes1 = new ArrayList<AssociationNode>();
    	nu.addAll(node.nodesUnmapped);
    	nodes1.addAll(G1Nodes);
    	double h = 0;
    	int left = G1Nodes.size() - index;
    	if(left <= nu.size()){
    		h += 2*(nu.size() - left);
    		for(int i = index; i < G1Nodes.size(); i ++){
        		double min = 2;
        		int indexj = -1;
        		for(int j = 0; j < nu.size(); j ++){
        			double cost = NodeEditCost(G1Nodes.get(i).getId(),nu.get(j).getId());
        			if(cost < min){
        				min = cost;
        				indexj = j;
        			}
        			
        		}
        		if(nu.size() != 0){
        			h += min;
            		nu.remove(nu.get(indexj));
        		}
        	}
    	}
    	else{
    		h += 2*(left - nu.size());
    		for(int i = 0; i < nu.size(); i ++){
    			double min = 2;
    			int indexj = -1;
    			for(int j = index; j < nodes1.size(); j ++){
    				double cost = NodeEditCost(nodes1.get(j).getId(), nu.get(i).getId());
    				if(cost < min){
    					min = cost;
    					indexj = j;
    				}
    			}
    			if(nodes1.size() > index){
    				h += min;
        			nodes1.remove(indexj);
    			}
    		}
    	}
    	return h;
    }
    private Node Initialize(ArrayList<AssociationNode> nodes1, ArrayList<AssociationNode> nodes2){
    	ArrayList<Mapping> initList = new ArrayList<Mapping>();
    	int size = nodes1.size();
    	for(int i = 0; i < size; i ++){
    		Mapping map = new Mapping(nodes1.get(i), nodes2.get(i));
    		initList.add(map);
    	}
        Node ini = new Node(initList);
        double GE = 0;
        for(int i = 0; i < size - 1; i ++){
        	for(int j = i + 1; j < size; j ++){
        		GE += EdgeEditCost(FindEdge(nodes1.get(i), nodes1.get(j)),FindEdge(nodes2.get(i), nodes2.get(j)));
        	}
        }
        ini.GE = GE;
        return ini;
    }
   
    private double NodeEditCost(int id1, int id2){//id2 = 0 indicates delete node id1
    	if(id1 == id2)
    		return 0;
    	else if(id2 == 0)
    		return 1;
    	else{
    		if(DEBUG_MODE){
    			System.out.println("ID: " + id1 + " ID2: " + id2 + " " + classDistance.getDistance(id1, id2));
    		}
    			
    		return classDistance.getDistance(id1, id2);
    	}
    }
    private double EdgeEditCost(int id1, int id2){
    	if(id1 == id2)
    		return 0;
    	else
    		return 1;
    }
    private double CalculateEdge(Node parent, Mapping map){//call it to calculate the GE when extends a node in openlist
       
    	double GE = parent.GE;
    	ArrayList<Mapping> parentel = parent.editlist;
    	Iterator<Mapping> it = parentel.iterator();
    	for(;it.hasNext();){
    		Mapping m = it.next();
    		GE += EdgeEditCost(FindEdge(m.fromG1,map.fromG1),FindEdge(m.fromG2,map.fromG2));
    	}
    	
    	return GE;
    }
    /**
     * @param n1
     * @param n2
     * @return the edge between n1 and n2, if there is no edge, returns 0
     */
    private int FindEdge(AssociationNode n1, AssociationNode n2){
    	if(n1 != null && n2 != null){
    		if(n1.getFather() != null && n1.getFather() == n2)//n1 is n2's child
        		return n1.getRelation();
        	if(n2.getFather() != null && n1 == n2.getFather())//n2 is n1's child
        		return -1*n2.getRelation();        	
        	return 0;
    	}
    	else return 0;
    }
    public Node MinF(List<Node> list)  
    {         
        Node min= list.get(0);  
        for(int i = 0; i<list.size();i++)  
        {  
            if(list.get(i).F<=min.F)  
            {  
                min = list.get(i);  
            }  
        }  
        return min;  
    }  
}  