/*
 * Licensed to meast-summarize under one or more contributor
 * license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright
 * ownership. IlluSnip licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package ws.meast.summarization.verbalizeimp;

/**
 * 
 * @author Yue Liang
 * This class is the default implementation of the verbalizer.
 * 
 */
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import name.dxliu.associations.AssociationNode;
import ws.meast.summarization.verbalize.AssociationTreeVerbalizer;
import ws.meast.summarization.verbalize.Dict;

public class AssociationTreeVerbalizerImp implements AssociationTreeVerbalizer{
	private static int MAX_CONTENT_LIMIT = 1000;

	public VerbalizedVertex[] vertice = new VerbalizedVertex[MAX_CONTENT_LIMIT];
	public VerbalizedEdge[] edges = new VerbalizedEdge[MAX_CONTENT_LIMIT];
	public List<String[]> finaltext = new ArrayList<String[]>();
	public EdgeComparator mc = new EdgeComparator();
	public Map<Integer, Integer> vertexTag = new HashMap<>();
	public Map<String, Integer> edgeTag = new HashMap<>();
	public int vertexNum = 0, edgeNum = 0;
	public String[] relatival = {"Person", "Writer", "Scientist", "Office Holder"};
	int[] flag = new int[MAX_CONTENT_LIMIT];
	int max;

	private String clause(final int vertex1, int vertex2, int vertex3) {
		String sentence = "";

		String subject, predicate1, object1, predicate2, object2;
		VerbalizedEdge e1 = null, e2 = null;
		for (VerbalizedEdge e:vertice[vertex1].next)
			if (e.to == vertex2) e1 = e;
		for (VerbalizedEdge e:vertice[vertex2].next)
			if (e.to == vertex3) e2 = e;
		
		subject = vertice[vertex1].label;
		predicate1 = e1.verb + e1.comp;
		object1 = vertice[e1.to].label; 
		
		sentence = subject + predicate1 + object1;
		sentence += getRelatival(e1.to);
		
		predicate2 = e2.verb + e2.comp;
		object2 = vertice[e2.to].label;
		sentence += predicate2 + object2;

		return sentence + ". ";
	}
	
	private String beshift(final int root, List<VerbalizedEdge> next, int axis) {
		String sentenceText = "";
		new ArrayList<String>();
		int i = 0;

		sentenceText += next.get(i).verb + next.get(i).comp + vertice[next.get(i).to].label;
		while (i < next.size() - 1) {
			i++;
			VerbalizedEdge now = next.get(i), last = next.get(i-1);
			
			if (now.verb.equals(last.verb)) {
				if (now.comp.equals(last.comp)) {
					if (i < next.size() - 1) {
						VerbalizedEdge nex = next.get(i+1);
						if (now.comp.equals(nex.comp)) 
							sentenceText += ", " + vertice[now.to].label;
						else sentenceText += " and " + vertice[now.to].label;
					}
					else {
						sentenceText += " and " + vertice[now.to].label;
					}
				}
				else {
					sentenceText += " and" + now.comp + vertice[now.to].label;
				}
			}
			else {
				if (i == next.size() - 1) sentenceText += " and";
				else sentenceText += ",";
				if (!now.verb.equals(last.verb)) sentenceText += now.verb;
				sentenceText += next.get(i).comp + vertice[now.to].label;
			}
		}
		return sentenceText + ". ";
	}

	private String triple(final int vertex, VerbalizedEdge edge) {
		String subject, predicate, object;
		subject = vertice[vertex].label;
		predicate = edge.verb + edge.comp;
		object = vertice[edge.to].label;
		
		return subject + predicate + object;
	}
	
	private String diameter1(final int center) {
		return triple(center, vertice[center].next.get(0));
	}
	
	private String diameter2(final int center) {
		String sentence = "";
		//System.out.println("center = " +center);
		if (vertice[center].next.size() == 1) {
			//center is a leaf
			VerbalizedEdge edge = vertice[center].next.get(0);
			if (vertice[edge.to].next.size() == 2) {
				//center->b->c clause
				int b = edge.to, c = -1;
				for (VerbalizedEdge e:vertice[b].next)
					if (e.to != b) c = e.to;
				sentence = clause(center, b, c);
			}
			else {
				//a->b->(c1,c2...)
				int b = vertice[center].next.get(0).to;
				List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
				for (VerbalizedEdge edge1:vertice[b].next) {
					if (edge1.to != center)
						next.add(edge1);
				}
				Collections.sort(next, mc);
				sentence = triple(center, vertice[center].next.get(0))+ getRelatival(vertice[center].next.get(0).to) + beshift(b, next, -1);
			}
		}
		else {
			//center is not a leaf
			List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
			for (VerbalizedEdge edge:vertice[center].next) {
				next.add(edge);
			}
			Collections.sort(next, mc);
			sentence = vertice[center].label + beshift(center, next, -1);
		}
		return sentence;
	}

	private String getRelatival(int to) {
		String temp = vertice[to].label;
		for (String str: relatival)
			if (temp.startsWith(str)) return " who";
		return " which";
	}

	private String diameter3(final int center) {
		String sentence = "";
		int b1 = -1, b2 = -1;
		//decide b1, b2;
		if (vertice[center].next.size() == 1) {
			// center is a leaf
			b1 = vertice[center].next.get(0).to;
		}
		else {
			//center is not a leave
			b1 = center;
		}
		for (VerbalizedEdge edge:vertice[b1].next) {
			if (vertice[edge.to].next.size() != 1) {
				b2 = edge.to;
				break;
			}
		}
		//System.out.println(vertice[b1].label+ vertice[b2].label);
	
		// the side contains the center
		if (vertice[center].next.size() == 1) {
			//center is a leaf
			if (vertice[b1].next.size() == 2) {
				//a->b1->b2 clause
				sentence = clause(center, b1, b2);
			//	System.out.println("con1");
			}
			else {
				//System.out.println("con2");
				//a->b1->(b2,a2,a3...) b2 is the axis
				List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
				for (VerbalizedEdge edge:vertice[b1].next) {
					if (edge.to == b2) {
						VerbalizedEdge e = edge;
						e.pri = 10;
						next.add(e);
					}
					else if (edge.to != center)
						next.add(edge);
				}
				Collections.sort(next, mc);
				sentence = triple(center, vertice[center].next.get(0)) + " which" + beshift(vertice[center].next.get(0).to, next, b2);
			}
		}
		else {

			//System.out.println("con3");
			//center == b1
			List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
			for (VerbalizedEdge edge:vertice[b1].next) {
				if (edge.to == b2) {
					VerbalizedEdge e = edge;
					e.pri = 10;
				}
				next.add(edge);
			}
			Collections.sort(next, mc);
			sentence = vertice[b1].label + beshift(b1, next, b2);
		}

		//the other side
		//b2->(ak, ak+1...)
		List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
		for (VerbalizedEdge edge:vertice[b2].next) 
			if (edge.to != b1)
			next.add(edge);
		Collections.sort(next, mc);
		sentence += vertice[b2].label + beshift(b2, next, -1);
		return sentence;
	}
	
	private String diameter4(int center) {
		String sentence = "";
		//4.1 find main
		int main = -1;
		for (int i = 1; i <= vertexNum; i++) {
			int count = 0;
			for (VerbalizedEdge edge:vertice[i].next) {
				int son = edge.to;
				if (vertice[son].next.size() >= 2) count ++;
			}
			if (count >= 2) {
				main = i;
				break;
			}
		}
		//System.out.println(vertice[main].uri);

		//4.2 deal with subtree which has diameter 2
		if (main == center) {
			//4.2.1 main == center
			//meanwhile, furthermore, moreover.
			for (VerbalizedEdge edge:vertice[main].next) {
				int b = edge.to;
				if (vertice[b].next.size() == 2) {
					//center->b->c
					int c = -1;
					for (VerbalizedEdge e:vertice[b].next)
						if (e.to != b) c = e.to;
					sentence += clause(center, b, c);
				}
				else if (vertice[b].next.size() >= 2) {
					//center->b->(c1,...ck)
					List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
					for (VerbalizedEdge e:vertice[b].next) {
						if (e.to != center)
							next.add(e);
					}
					Collections.sort(next, mc);
					sentence += triple(center, edge) + " which " + beshift(b, next, -1);
				}
			}
		}
		else {
			//4.2.2 main != center
			//the subtree contain center
			int b = -1;
			if (vertice[center].next.size() == 1) {
				//4.2.2.1 center is a leaf
				//center->(b...)->main
				b = vertice[center].next.get(0).to;
				if (vertice[b].next.size() == 2) {
					//center->b->main clause
					sentence += clause(center, b, main);
				}
				else {
					//center->(b...)->main, main is the axis					
					List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
					for (VerbalizedEdge edge:vertice[b].next) {
						if (edge.to == main) {
							VerbalizedEdge e = edge;
							e.pri = 10;
							next.add(e);
						}
						else if (edge.to != center)
							next.add(edge);
					}
					Collections.sort(next, mc);
					sentence += triple(center, vertice[center].next.get(0)) + " which" + beshift(b, next, main);
				}
			}
			else {
				//4.2.2.2 center is not a leaf
				//center->(...,main), main is the axisList<Edge> next = new ArrayList<Edge>();
				b = center;
				List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
				for (VerbalizedEdge edge:vertice[center].next) {
					if (edge.to == main) {
						VerbalizedEdge e = edge;
						e.pri = 10;
						next.add(e);
					}
					else next.add(edge);
				}
				Collections.sort(next, mc);
			//	System.out.println(vertice[center].uri);
				Collections.sort(next, mc);
				sentence += vertice[center].label + beshift(center, next, main);
			}
			//the rest subtree
			//main->(b...)
			for (VerbalizedEdge edge:vertice[main].next) 
				if (edge.to != b && vertice[edge.to].next.size() >= 2) { //exclude the subtree contains the center
					int b1 = edge.to;
			//		System.out.println(edge.to);
					if (vertice[b1].next.size() == 2) {
						//main->b1->c
						int c = -1;
						for (VerbalizedEdge e:vertice[b1].next)
							if (e.to != main) c = e.to;
						sentence += clause(main, b1, c);
					}
					else {
						//center->b1->(c1,...ck)
						List<VerbalizedEdge> next = new ArrayList<VerbalizedEdge>();
						for (VerbalizedEdge e:vertice[b1].next) {
							if (e.to != main)
								next.add(e);
						}
						Collections.sort(next, mc);
						sentence += triple(main, edge) + " which" + beshift(b1, next, -1);
						//sentence += vertice[main].label + beshift(b1, next, -1);
					}
				}
		}
		
		//4.3 deal with subtree which has diameter 1
		List<VerbalizedEdge> remain = new ArrayList<VerbalizedEdge>();
		for (VerbalizedEdge e:vertice[main].next) 
			if (vertice[e.to].next.size() == 1) {
				remain.add(e);
		//		System.out.println(e.uri);
			}
		Collections.sort(remain, mc);
		if (remain.size() > 0)
			sentence += vertice[main].label + beshift(main, remain, -1);
		return sentence;
	}


	
	void dfs(int k, int len) {
		if (len > max) max = len;
		for (VerbalizedEdge e:vertice[k].next) 
			if (flag[e.to] == 0) {
				flag[e.to] = 1;
				dfs(e.to, len+1);
				flag[e.to] = 0;
			}
	}
	
	int getDiameter() {
		max = 0;
		for (int i = 1; i <= vertexNum; i++) {
			for (int j = 1; j <= vertexNum; j++) flag[j] = 0;
			flag[i] = 1;
			dfs(i, 0);
		}
		return max;
	}
	
	private void init() {
		vertice = new VerbalizedVertex[MAX_CONTENT_LIMIT];
		edges = new VerbalizedEdge[MAX_CONTENT_LIMIT];
		finaltext = new ArrayList<String[]>();
	    mc = new EdgeComparator();
		vertexTag = new HashMap<>();
		edgeTag = new HashMap<>();
		vertexNum = 0;
		edgeNum = 0;
	}
	
	private void code2tree(Dict dict,List<Integer> code) {
		//rebuild the association tree from the canonical code
		
		Stack<Integer> node = new Stack<Integer>();
		vertexNum = 0;
		int i = 0;
		
		while (i < code.size()) {
			vertexNum++;
			vertexTag.put(code.get(i), vertexNum);			
			while (i+1 < code.size() && code.get(i+1) == AssociationNode.$) 
				i++;
			i += 2;
		}
							
		node.add(code.get(0));
		int j = 1;
		while (j < code.size()) {
			int s = node.peek();
			int p = code.get(j);
			int absp = p > 0 ? p : -p;
			int o = code.get(j+1);
			
			node.add(o);
			int sub = vertexTag.get(s), obj = vertexTag.get(o);

			edges[++edgeNum] = new VerbalizedEdge(dict.getLabel(absp), p, obj, getDirection(p));
			
			if (vertice[sub] == null) 
				vertice[sub] = new VerbalizedVertex(dict.getLabel(s), s);
			vertice[sub].next.add(edges[edgeNum]);	
	
			edges[++edgeNum] = new VerbalizedEdge(dict.getLabel(absp), -p, sub, getDirection(-p));
			
			if (vertice[obj] == null)
				vertice[obj] = new VerbalizedVertex(dict.getLabel(o), o);
			vertice[obj].next.add(edges[edgeNum]);
			
			j += 2;
			while (j < code.size() && code.get(j) == AssociationNode.$) { 
				j++;
				node.pop();
			}		
		}
	}
	
	private int getDirection(int p) {
		return p >= 0 ? 1 : -1;
	}
	
	@Override
	public String generateText(final Dict dict,final List<Integer> code, final int c) {
		init();		
		code2tree(dict,code);		
		String text = "";		
		int center = 0;
		
		int diameter = getDiameter();
		
		for (int i = 1; i <= vertexNum; i++) {
			if (vertice[i].label.equals(dict.getLabel(c)))
				center = vertice[i].id; 
		}
		//System.out.println("center "+ center);
		
		if (diameter == 1) 
			text =  diameter1(vertexTag.get(center));
		if (diameter == 2) {
			text = diameter2(vertexTag.get(center));
		}
		if (diameter == 3) {
			text = diameter3(vertexTag.get(center));
		}
		if (diameter == 4) {
			text = diameter4(vertexTag.get(center));
		}
		//System.out.println(text);
		return text;
	}
}
