/*
 * Licensed to meast-summarize under one or more contributor
 * license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright
 * ownership. IlluSnip licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package ws.meast.summarization.verbalizeimp;
/**
 * 
 * @author Yue Liang
 * 
 */

import java.util.Properties;

import edu.stanford.nlp.simple.Sentence;  
    
public class VerbalizedEdge {
	private static int MAX_CONTENT_LIMIT = 1000;
	
	public String[] list = new String[MAX_CONTENT_LIMIT];
	public String[] PP = {"By", "In", "After"};
	public String uri, label, prop, superprop, verb, comp;
	public int to, relation, type, pri;
	public Properties props = new Properties();      
	
	public int getPri() {
			return this.pri;
	}

	private void tokenlize(String label) {
		String[] words = label.split("(?=\\p{Upper})");
		for (int i = 0; i < words.length; i++) {
			words[i] = words[i].toLowerCase();
		}

	   	String pos = new Sentence(words[0]).posTag(0);
	   	String lemma = new Sentence(words[0]).lemma(0);
	    //	System.out.println(words[0] + lemma + pos);

	   	if ((pos.equals("VBN") || pos.equals("VBD")) && words.length > 1) {
			this.verb = " is";
			this.comp = " " + words[0] + " " + words[1] + " ";
	   		this.pri = 1;
	    }else if ((pos.equals("VBN") || pos.equals("VBD"))) {
	    	this.verb = " " + lemma + " ";
	   		this.comp = "";
	   		this.pri = 1;
	    }else if (words[0].equals("is")) {
	       	this.verb = " is";
	       	this.comp = " the " + words[1] + " of ";
	       	this.pri = 3;
	    }else if (words[0].equals("has")) {
	       	this.verb = " has";
	       	this.comp = " " + words[1] + " ";
	       	this.pri = 2;
	    }else {
			this.verb = " has";
			this.comp = " ";
			for (String word:words) {
				this.comp += word + " ";
			}
   		 	this.pri = 2;
        }
	}


	private void reverlize(String label) {
		String[] words = label.split("(?=\\p{Upper})");
		for (int i = 0; i < words.length; i++) {
			words[i] = words[i].toLowerCase();
		}
		String pos = new Sentence(words[0]).posTag(0);
		String lemma = new Sentence(words[0]).lemma(0);
		//System.out.println(words[0] + lemma + pos);
	    	
		if ((pos.equals("VBN") || pos.equals("VBD")) && words.length > 1) {
			this.verb = " " + lemma + " ";
			this.comp = "";
			this.pri = 1;
	    }else if ((pos.equals("VBN") || pos.equals("VBD"))) {
	    	this.verb = " is";
	    	this.comp = " " + words[0] + " by ";
	    	this.pri = 1;
	    }else if (pos.equals("VB")) {
	    	this.verb = " is";
	    	this.comp = " " + words[0] + " by ";
	    	this.pri = 1;
	    }else if (words[0].equals("is")) {
	    	this.verb = " has";
	    	this.comp = " " + words[1] + " ";
	    	this.pri = 2;
	    }else if (words[0].equals("has")) {
	    	this.verb = " is";
	    	this.comp = " the " + words[1] + " of ";
	    	this.pri = 3;
	    }else {
	    	this.verb = " is";
	    	this.comp = " the ";
	    	for (String word:words) {
	    		this.comp += word + " ";
			}
			this.comp += "of ";
			this.pri = 3;
	   	}			
	}

	public VerbalizedEdge(String label, int id, int to, int relation) {
		this.label = label;
		this.to = to;
		this.relation = relation;
		this.prop = uri;
		
		if (relation == 1) {
			tokenlize(label);
		}else if (relation == -1) {
			reverlize(label);
		}	
	}
}